//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/RealModel.cpp
//! @brief     Implements class RealModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/RealModel.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Model/SessionXML.h"

RealModel::RealModel(QObject* parent)
    : SessionModel(GUI::Session::XML::RealDataModelTag, parent)
{
    setObjectName(GUI::Session::XML::RealDataModelTag);

    connect(this, &SessionModel::rowsInserted, this, &RealModel::onRowsChange);
    connect(this, &SessionModel::rowsRemoved, this, &RealModel::onRowsChange);
}

QVector<SessionItem*> RealModel::nonXMLItems() const
{
    QVector<SessionItem*> result;

    for (auto* realData : topItems<RealItem>()) {
        if (auto* intensityItem = realData->dataItem())
            result.push_back(intensityItem);
        if (auto* native_data_item = realData->nativeData())
            result.push_back(native_data_item);
    }

    return result;
}

void RealModel::readFrom(QXmlStreamReader* reader, MessageService* messageService /*= 0*/)
{
    // do not send added-notifications until completely read - otherwise partially
    // initialized items will be notified
    disconnect(this, &SessionModel::rowsInserted, this, &RealModel::onRowsChange);

    SessionModel::readFrom(reader, messageService);

    connect(this, &SessionModel::rowsInserted, this, &RealModel::onRowsChange);

    if (!realItems().isEmpty())
        emit realDataAddedOrRemoved();
}

RealItem* RealModel::insertSpecularDataItem()
{
    auto* r = insertItem<RealItem>();
    r->initAsSpecularItem();
    return r;
}

RealItem* RealModel::insertIntensityDataItem()
{
    auto* r = insertItem<RealItem>();
    r->initAsIntensityItem();
    return r;
}

RealItem* RealModel::insertDataItem(int rank)
{
    return (rank == 1) ? insertSpecularDataItem() : insertIntensityDataItem();
}

QVector<RealItem*> RealModel::realItems() const
{
    return topItems<RealItem>();
}


QVector<RealItem*> RealModel::realItems(int rank) const
{
    return rank == 1 ? realItems1D() : realItems2D();
}

QVector<RealItem*> RealModel::realItems1D() const
{
    return topItems<RealItem>([](const RealItem& item) { return item.isSpecularData(); });
}

QVector<RealItem*> RealModel::realItems2D() const
{
    return topItems<RealItem>([](const RealItem& item) { return !item.isSpecularData(); });
}

void RealModel::remove(RealItem* item)
{
    removeItem(item);
}

QStringList RealModel::realDataNames() const
{
    QStringList result;
    for (auto item : realItems())
        result.append(item->dataName());
    return result;
}

void RealModel::onRowsChange(const QModelIndex& parent, int, int)
{
    // valid parent means not a data (which is top level item) but something below
    if (parent.isValid())
        return;

    emit realDataAddedOrRemoved();
}
