//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Param/Node/INode.cpp
//! @brief     Implements interface INode.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Param/Node/INode.h"
#include "Base/Util/Assert.h"
#include "Fit/Param/RealLimits.h"
#include <algorithm>
#include <iostream>
#include <utility>

INode::INode(std::vector<double> PValues)
    : m_P(std::move(PValues))
{
}

void INode::checkNodeArgs() const
{
    size_t nP = m_P.size();
    if (parDefs().size() != nP) {
        std::cerr << "BUG in class " << className() << std::endl;
        std::cerr << "#m_P = " << nP << std::endl;
        std::cerr << "#PDf = " << parDefs().size() << std::endl;
        for (const ParaMeta& pm : parDefs())
            std::cerr << "  PDf: " << pm.name << std::endl;
        ASSERT(0);
    }
    ASSERT(parDefs().size() == nP);
    for (size_t i = 0; i < nP; ++i) {
        const ParaMeta pm = parDefs()[i];

        RealLimits limits = RealLimits::limitless();
        if (pm.vMin == -INF) {
            ASSERT(pm.vMax == +INF);
            // nothing to do
        } else if (pm.vMax == +INF) {
            ASSERT(pm.vMin == 0);
            limits = RealLimits::nonnegative();
        } else {
            limits = RealLimits::limited(pm.vMin, pm.vMax);
        }
        limits.check(pm.name, m_P[i]);
    }
}

std::vector<const INode*> INode::nodeChildren() const
{
    return {};
}

std::vector<const INode*> INode::nodeOffspring() const
{
    std::vector<const INode*> result;
    result.push_back(this);
    for (const auto* child : nodeChildren()) {
        for (const auto* p : child->nodeOffspring())
            result.push_back(p);
    }
    return result;
}
