//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/Slicer.cpp
//! @brief     Implements function Compute::Slicing::sliceFormFactor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Processed/Slicer.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"
#include "Base/Util/Assert.h"
#include "Base/Vector/RotMatrix.h"
#include "Resample/Particle/ReCompound.h"
#include "Resample/Particle/ReCoreShell.h"
#include "Resample/Particle/ReMesocrystal.h"
#include "Resample/Particle/ReParticle.h"
#include "Resample/Processed/ParticleInSlice.h"
#include "Resample/Slice/ZLimits.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Material/MaterialFactoryFuncs.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleComposition.h"
#include "Sample/Particle/ParticleCoreShell.h"
#include "Sample/Scattering/Rotations.h"

namespace {

//! Nested structure that holds slicing effects on position and removed parts.
struct SlicingEffects {
    R3 position;
    double dz_bottom;
    double dz_top;
};

SlicingEffects computeSlicingEffects(ZLimits limits, const R3& position, double height)
{
    R3 new_position(position);
    double z_bottom = position.z();
    double z_top = position.z() + height;
    double dz_top = std::isinf(limits.zTop()) ? -1 : z_top - limits.zTop();
    double dz_bottom = std::isinf(limits.zBottom()) ? -1 : limits.zBottom() - z_bottom;
    ASSERT(dz_top >= 0 || dz_bottom >= 0);
    ASSERT(dz_bottom <= height);
    ASSERT(dz_top <= height);
    if (dz_bottom < 0)
        dz_bottom = 0;
    if (dz_top < 0)
        dz_top = 0;
    if (dz_bottom > 0)
        new_position.setZ(limits.zBottom());
    return {new_position, dz_bottom, dz_top};
}

bool shapeIsContainedInLimits(const IFormFactor& formfactor, ZLimits limits,
                              const IRotation* rotation, R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return limits.zBottom() <= zbottom && ztop <= limits.zTop();
}

bool shapeOutsideLimits(const IFormFactor& formfactor, ZLimits limits, const IRotation* rotation,
                        R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return ztop <= limits.zBottom() || zbottom >= limits.zTop();
}

ReParticle* createParticleSlice(const IFormFactor* ff, ZLimits limits, const IRotation* rot,
                                R3 translation)
{
    if (shapeOutsideLimits(*ff, limits, rot, translation))
        return nullptr;

    if (shapeIsContainedInLimits(*ff, limits, rot, translation))
        return ReParticle::createTransformedFormFactor(*ff, rot, translation);

    if (!ff->canSliceAnalytically(rot))
        throw std::runtime_error("Slicing of " + ff->className()
                                 + " not supported for the given rotation");

    if (const auto* f = dynamic_cast<const Pyramid2*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dbase_edge = 2 * effects.dz_bottom * Math::cot(f->alpha());
        Pyramid2 slicedff(f->length() - dbase_edge, f->width() - dbase_edge,
                          f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Box*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        Box slicedff(f->length(), f->width(), f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Cone*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dradius = effects.dz_bottom * Math::cot(f->alpha());
        Cone slicedff(f->radius() - dradius, f->height() - effects.dz_bottom - effects.dz_top,
                      f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid6*>(ff)) {
        const SlicingEffects effects = computeSlicingEffects(limits, translation, f->height());
        double dbase_edge = effects.dz_bottom * Math::cot(f->alpha());
        Pyramid6 slicedff(f->baseEdge() - dbase_edge,
                          f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Bipyramid4*>(ff)) {
        const SlicingEffects effects =
            computeSlicingEffects(limits, translation, f->height() * (1 + f->heightRatio()));
        if (effects.dz_bottom > f->height()) {
            double dbase_edge = 2 * (effects.dz_bottom - f->height()) * Math::cot(f->alpha());
            Pyramid4 slicedff(f->length() - dbase_edge,
                              f->height() * (1 + f->heightRatio()) - effects.dz_bottom
                                  - effects.dz_top,
                              f->alpha());
            return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);
        }
        if (effects.dz_top > f->heightRatio() * f->height()) {
            double dbase_edge = 2 * (f->height() - effects.dz_bottom) * Math::cot(f->alpha());
            Pyramid4 slicedff(f->length() - dbase_edge,
                              f->height() * (1 + f->heightRatio()) - effects.dz_bottom
                                  - effects.dz_top,
                              M_PI - f->alpha());
            return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);
        }
        Bipyramid4 slicedff(f->length(), f->height() - effects.dz_bottom,
                            f->heightRatio() * f->height() - effects.dz_top, f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Cylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        Cylinder slicedff(f->radius(), f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const EllipsoidalCylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        EllipsoidalCylinder slicedff(f->radiusX(), f->radiusY(),
                                     f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const HorizontalCylinder*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, 2 * f->radius());
        HorizontalCylinder slicedff(f->radius(), f->length(), -f->radius() + effects.dz_bottom,
                                    f->radius() - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Sphere*>(ff)) {
        R3 new_translation = translation;
        if (rot) {
            R3 center(0.0, 0.0, f->radius());
            new_translation += rot->transformed(center) - center;
        }
        const double h = 2 * f->radius();
        auto effects = computeSlicingEffects(limits, translation, h);
        TruncatedSphere slicedff(f->radius(), h - effects.dz_bottom, effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, nullptr, effects.position);

    } else if (const auto* f = dynamic_cast<const Spheroid*>(ff)) {
        double flattening = f->height() / (2.0 * f->radius());
        auto effects = computeSlicingEffects(limits, translation, f->height());
        TruncatedSpheroid slicedff(f->radius(), f->height() - effects.dz_bottom, flattening,
                                   effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const LongBoxGauss*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        LongBoxGauss slicedff(f->length(), f->width(),
                              f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const LongBoxLorentz*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        LongBoxLorentz slicedff(f->length(), f->width(),
                                f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Prism3*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        Prism3 slicedff(f->baseEdge(), f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Prism6*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        Prism6 slicedff(f->baseEdge(), f->height() - effects.dz_bottom - effects.dz_top);
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid4*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        double dbaseEdge = 2 * effects.dz_bottom * Math::cot(f->alpha());
        Pyramid4 slicedff(f->baseEdge() - dbaseEdge,
                          f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const Pyramid3*>(ff)) {
        auto effects = computeSlicingEffects(limits, translation, f->height());
        double dbaseEdge = 2 * sqrt(3) * effects.dz_bottom * Math::cot(f->alpha());
        Pyramid3 slicedff(f->baseEdge() - dbaseEdge,
                          f->height() - effects.dz_bottom - effects.dz_top, f->alpha());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const TruncatedSphere*>(ff)) {
        double height = f->height() - f->removedTop();
        auto effects = computeSlicingEffects(limits, translation, height);
        TruncatedSphere slicedff(f->radius(), f->height() - effects.dz_bottom,
                                 effects.dz_top + f->removedTop());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else if (const auto* f = dynamic_cast<const TruncatedSpheroid*>(ff)) {
        double height = f->height() - f->removedTop();
        auto effects = computeSlicingEffects(limits, translation, height);
        TruncatedSpheroid slicedff(f->radius(), f->height() - effects.dz_bottom,
                                   f->heightFlattening(), effects.dz_top + f->removedTop());
        return ReParticle::createTransformedFormFactor(slicedff, rot, effects.position);

    } else
        throw std::runtime_error("Slicing of " + ff->className() + " not supported");
}

//! Recursively processes the basis of a mesocrystal.
IReParticle* processBasis(const IParticle* basis)
{
    if (const auto* b = dynamic_cast<const ParticleComposition*>(basis)) {
        const auto& particles = b->decompose();
        ASSERT(!particles.empty());
        auto* result = new ReCompound;
        for (const auto& particle : particles) {
            std::unique_ptr<IReParticle> re(processBasis(particle.get()));
            result->addFormFactor(*re);
        }
        return result;
    }
    return Compute::Slicing::createParticleInSlice(basis, {}).particleSlice.release();
}

} // namespace


ParticleInSlice Compute::Slicing::createParticleInSlice(const IParticle* particle,
                                                        const ZLimits& limits)
{
    if (const auto* p = dynamic_cast<const Particle*>(particle)) {
        ASSERT(p->formfactor_at_bottom());
        std::unique_ptr<ReParticle> particleSlice(createParticleSlice(
            p->formfactor_at_bottom(), limits, p->rotation(), p->particlePosition()));
        if (!particleSlice)
            return {};
        double volume = particleSlice->volume();
        Material transformed_material(
            p->rotation() ? p->material()->rotatedMaterial(p->rotation()->rotMatrix())
                          : *p->material());
        particleSlice->setMaterial(transformed_material);
        return {std::move(particleSlice), {{{volume, transformed_material}}}};

    } else if (const auto* p = dynamic_cast<const ParticleCoreShell*>(particle)) {
        const Particle* core = p->coreParticle();
        const Particle* shell = p->shellParticle();
        ASSERT(core && shell);

        // core
        std::unique_ptr<Particle> P_core(core->clone());
        if (p->rotation())
            P_core->rotate(*p->rotation());
        P_core->translate(p->particlePosition());
        ParticleInSlice sliced_core = createParticleInSlice(P_core.get(), limits);

        // shell
        std::unique_ptr<Particle> P_shell(shell->clone());
        if (p->rotation())
            P_shell->rotate(*p->rotation());
        P_shell->translate(p->particlePosition());
        ParticleInSlice sliced_shell = createParticleInSlice(P_shell.get(), limits);
        if (!sliced_shell.particleSlice)
            return {};

        ParticleInSlice result;
        // if core out of limits, return sliced shell
        if (!sliced_core.particleSlice) {
            result.particleSlice.reset(std::move(sliced_shell.particleSlice.get()));
            result.admixtures.push_back(sliced_shell.admixtures.back());
            return result;
        }

        // set core ambient material
        if (sliced_shell.admixtures.size() != 1)
            return {};
        auto shell_material = sliced_shell.admixtures[0].material;
        sliced_core.particleSlice->setAmbientMaterial(shell_material);

        // construct sliced particle
        sliced_shell.admixtures.back().fraction -= sliced_core.admixtures.back().fraction;
        result.particleSlice.reset(new ReCoreShell(sliced_core.particleSlice.release(),
                                                   sliced_shell.particleSlice.release()));
        result.admixtures.push_back(sliced_core.admixtures.back());
        result.admixtures.push_back(sliced_shell.admixtures.back());

        return result;

    } else if (dynamic_cast<const ParticleComposition*>(particle)) {
        throw std::runtime_error("ParticleComposition does not yet support slicing");

    } else if (const auto* p = dynamic_cast<const MesoCrystal*>(particle)) {
        const Crystal* crystal = &p->particleStructure();
        const IFormFactor* meso_formfactor = p->outerShape();
        ASSERT(crystal && meso_formfactor);

        double unit_cell_volume = crystal->lattice()->unitCellVolume();
        if (unit_cell_volume <= 0)
            return {};

        ParticleInSlice result;

        std::unique_ptr<ReParticle> new_shape(
            createParticleSlice(meso_formfactor, limits, p->rotation(), p->particlePosition()));

        const std::unique_ptr<Crystal> new_crystal(
            crystal->transformed(p->rotation(), p->particlePosition()));

        const std::unique_ptr<IReParticle> new_basis(processBasis(new_crystal->basis()));

        result.particleSlice.reset(new ReMesocrystal(*new_crystal->lattice(), *new_basis,
                                                     *new_shape, new_crystal->position_variance()));

        Admixtures& regions = result.admixtures;
        ASSERT(regions.empty());
        for (const auto& particle : crystal->basis()->decompose()) {
            ParticleInSlice pis = createParticleInSlice(particle.get(), {});
            regions.insert(regions.end(), pis.admixtures.begin(), pis.admixtures.end());
        }
        for (auto& region : regions)
            region.fraction *= meso_formfactor->volume() / unit_cell_volume;

        return result;

    } else
        ASSERT(0);
}

ZLimits Compute::Slicing::zSpan(const IParticle* particle)
{
    if (const auto* p = dynamic_cast<const ParticleComposition*>(particle)) {
        const auto subparticles = p->decompose();
        ASSERT(subparticles.size() > 0);
        ZLimits result = zSpan(subparticles[0].get());
        for (const auto& subparticle : subparticles)
            result = ZLimits::enclosingInterval(result, zSpan(subparticle.get()));
        return result;
    }
    const ParticleInSlice pis = createParticleInSlice(particle, ZLimits{});
    const double bot = pis.particleSlice->bottomZ(nullptr);
    const double top = pis.particleSlice->topZ(nullptr);
    return {bot, top};
}
