//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/HorizontalCylinder.cpp
//! @brief     Implements class HorizontalCylinder.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/HorizontalCylinder.h"
#include "Base/Math/Bessel.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"
#include "Base/Math/IntegratorGK.h"
#include "Sample/Shapes/DoubleEllipse.h"

HorizontalCylinder::HorizontalCylinder(const std::vector<double> P)
    : IFormFactor(P)
    , m_radius(m_P[0])
    , m_length(m_P[1])
    , m_slice_bottom(m_P[2])
    , m_slice_top(m_P[3])
{
    checkNodeArgs();
    // TODO improve!
    m_shape3D = std::make_unique<DoubleEllipseX>(-m_length / 2, m_radius, m_radius, m_slice_bottom,
                                                 m_slice_top, m_length / 2, m_radius, m_radius,
                                                 m_slice_bottom, m_slice_top);
}

HorizontalCylinder::HorizontalCylinder(double radius, double length, double slice_bottom,
                                       double slice_top)
    : HorizontalCylinder(std::vector<double>{radius, length, slice_bottom, slice_top})
{
}

HorizontalCylinder::HorizontalCylinder(double radius, double length)
    : HorizontalCylinder(radius, length, -radius, +radius)
{
}

complex_t HorizontalCylinder::formfactor_at_bottom(C3 q) const
{
    const double R = m_radius;
    const double L = m_length;

    const complex_t qL2 = q.x() * L / 2.;
    const complex_t qR = std::sqrt(q.y() * q.y() + q.z() * q.z()) * R;

    const complex_t axial_part = L * Math::sinc(qL2);
    complex_t radial_part;

    if (m_slice_bottom == -R && m_slice_top == +R)
        radial_part = M_TWOPI * R * R * Math::Bessel::J1c(qR) * exp_I(q.z() * R);

    else
        radial_part =
            2.
            * ComplexIntegrator().integrate(
                [=](double z) {
                    double y = sqrt(R * R - z * z);
                    return y * Math::sinc(q.y() * y) * exp_I(q.z() * (z - m_slice_bottom));
                },
                m_slice_bottom, m_slice_top);

    return radial_part * axial_part;
}
