//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Material/MaterialUtils.cpp
//! @brief     Implements functions in namespace MaterialUtils
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Material/MaterialUtils.h"
#include "Base/Const/PhysicalConstants.h"
#include "Base/Spin/SpinMatrix.h"

using PhysConsts::g_factor_n;
using PhysConsts::h_bar;
using PhysConsts::m_n;
using PhysConsts::mu_N;
// The factor 1e-18 is here to have unit: 1/T*nm^-2
constexpr double magnetic_prefactor = (m_n * g_factor_n * mu_N / h_bar / h_bar) * 1e-18;

namespace {

// Pauli matrices
const SpinMatrix Pauli_X(0, 1, 1, 0);
const SpinMatrix Pauli_Y(0, -I, I, 0);
const SpinMatrix Pauli_Z(1, 0, 0, -1);

} // namespace

template <typename T>
SpinMatrix MaterialUtils::MagnetizationCorrection(complex_t unit_factor, double magnetic_factor,
                                                  Vec3<T> polarization)
{
    SpinMatrix result = unit_factor * SpinMatrix::One()
                        + magnetic_factor
                              * (Pauli_X * polarization.x() + Pauli_Y * polarization.y()
                                 + Pauli_Z * polarization.z());
    return result;
}

// Prompt compilation for real and complex vectors:
template SpinMatrix MaterialUtils::MagnetizationCorrection(complex_t unit_factor,
                                                           double magnetic_factor, R3 polarization);
template SpinMatrix MaterialUtils::MagnetizationCorrection(complex_t unit_factor,
                                                           double magnetic_factor, C3 polarization);

complex_t MaterialUtils::ScalarReducedPotential(complex_t n, R3 k, double n_ref)
{
    return n * n - n_ref * n_ref * R3Util::sin2Theta(k);
}

SpinMatrix MaterialUtils::PolarizedReducedPotential(complex_t n, R3 b_field, R3 k, double n_ref)
{
    double factor = magnetic_prefactor / k.mag2();
    complex_t unit_factor = ScalarReducedPotential(n, k, n_ref);
    return MagnetizationCorrection(unit_factor, factor, b_field);
}

MATERIAL_TYPES MaterialUtils::checkMaterialTypes(const std::vector<const Material*>& materials)
{
    MATERIAL_TYPES result = MATERIAL_TYPES::RefractiveMaterial;
    bool isDefault = true;
    for (const Material* mat : materials) {
        if (isDefault) {
            result = mat->typeID();
            isDefault = mat->isDefaultMaterial();
            continue;
        }
        if (mat->typeID() != result && !mat->isDefaultMaterial())
            return MATERIAL_TYPES::InvalidMaterialType;
    }
    return result;
}
