//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ScatteringSimulation.cpp
//! @brief     Implements interface ISimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/ScatteringSimulation.h"
#include "Base/Axis/IAxis.h"
#include "Base/Util/Assert.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Detector/IDetector.h"
#include "Device/Histo/SimulationResult.h"
#include "Resample/Element/DiffuseElement.h"

ScatteringSimulation::ScatteringSimulation(const Beam& beam, const MultiLayer& sample,
                                           const IDetector& detector)
    : ISimulation2D(beam, sample, detector)
{
}

void ScatteringSimulation::prepareSimulation()
{
    if (m_detector->rank() != 2)
        throw std::runtime_error("ScatteringSimulation::prepareSimulation() "
                                 "-> Error. The detector was not properly configured.");
    ISimulation2D::prepareSimulation();
}

SimulationResult ScatteringSimulation::pack_result()
{
    auto* const coordsys = createCoordSystem();
    const std::unique_ptr<Datafield> data(m_detector->createDetectorIntensity(m_eles));
    return {*data, *coordsys};
}

#ifndef SWIG
ICoordSystem* ScatteringSimulation::createCoordSystem() const
{
    auto* const det2D = dynamic_cast<const IDetector*>(getDetector());
    ASSERT(det2D);
    return det2D->scatteringCoords(beam());
}
#endif

void ScatteringSimulation::initElementVector()
{
    m_eles = std::move(generateElements(beam()));

    if (m_cache.empty())
        m_cache.resize(m_eles.size(), 0.0);
}
