#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Model/RealModel.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/Path.h"
#include "GUI/View/Project/AutosaveController.h"
#include "GUI/View/Project/SaveService.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QSignalSpy>

class TestSaveService : public ::testing::Test {
protected:
    // helper method to modify something in a model
    void modify_models(ProjectDocument& doc)
    {
        auto* instrument = doc.collectedItems()->collectedItems().front();
        doc.instrumentsEditController()->setInstrumentName(instrument,
                                                           QUuid::createUuid().toString());
    }
    const int m_save_wait = 10000;
};

//! Testing AutosaveController. It watches ProjectDocument and sends autosaveRequest() when
//! number of document changes has been accumulated.

TEST_F(TestSaveService, autoSaveController)
{
    const QString projectDir("test_autoSaveController");
    UTest::GUI::create_dir(projectDir);

    const QString projectFileName(projectDir + "/document.pro");

    const int autosave_time(100);

    std::unique_ptr<ProjectDocument> document(new ProjectDocument);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");
    document->save(projectFileName);

    // setting up autosave
    AutosaveController autosave;
    autosave.setAutosaveTime(autosave_time);
    autosave.setDocument(document.get());

    // checking proposed autosave directory
    EXPECT_EQ(autosave.autosaveDir(), QString(projectDir + "/autosave"));
    EXPECT_EQ(autosave.autosaveName(), QString(projectDir + "/autosave/document.pro"));

    QSignalSpy spyAutosave(&autosave, SIGNAL(autosaveRequest()));

    // modify document once and check that autosave directory was created
    modify_models(*document);
    EXPECT_TRUE(document->isModified());
    EXPECT_TRUE(spyAutosave.wait(m_save_wait));
    EXPECT_EQ(spyAutosave.count(), 1);
    EXPECT_TRUE(QFile::exists(autosave.autosaveDir()));

    // saving document and checking that autosave is not triggered
    document->save(projectFileName);
    EXPECT_FALSE(document->isModified());
    EXPECT_EQ(spyAutosave.count(), 1);

    // modify several times and check than autosave was triggered only once
    for (size_t i = 0; i < 10; ++i)
        modify_models(*document);

    EXPECT_TRUE(spyAutosave.wait(m_save_wait));
    EXPECT_EQ(spyAutosave.count(), 2);

    // remove autosave dir
    autosave.removeAutosaveDir();
    EXPECT_FALSE(QFile::exists(autosave.autosaveDir()));
}

//! AutosaveController shouldn't trigger autosaveRequest() if document has no name.

TEST_F(TestSaveService, autoSaveControllerNewDocument)
{
    std::unique_ptr<ProjectDocument> document(new ProjectDocument);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");

    const int autosave_time(100);

    AutosaveController autosave;
    autosave.setAutosaveTime(autosave_time);
    autosave.setDocument(document.get());

    QSignalSpy spyAutosave(&autosave, SIGNAL(autosaveRequest()));

    modify_models(*document);
    EXPECT_FALSE(spyAutosave.wait(autosave_time * 2));
    EXPECT_EQ(spyAutosave.count(), 0);
}

//! Testing SaveService on simple documents (no heavy data).
//! SaveService should be able to save project file and send documentSaved() signal.

TEST_F(TestSaveService, saveService)
{
    const QString projectDir("test_saveService");
    UTest::GUI::create_dir(projectDir);
    const QString projectFileName(projectDir + "/document.pro");


    std::unique_ptr<ProjectDocument> document(new ProjectDocument);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");
    modify_models(*document);

    EXPECT_FALSE(QFile::exists(projectFileName));

    SaveService service;
    QSignalSpy spySaveService(&service, SIGNAL(projectSaved()));

    service.setDocument(document.get());
    service.save(projectFileName);

    EXPECT_EQ(spySaveService.count(), 1);
    EXPECT_TRUE(QFile::exists(projectFileName));
    EXPECT_FALSE(service.isSaving());

    // after save, document should be in non-modified state, project file name should be updated
    EXPECT_EQ(document->projectFileName(), projectFileName);
    EXPECT_FALSE(document->isModified());
}

//! Testing SaveService on simple documents (no heavy data).
//! SaveService should not be able to save project file; state of service has to be !isSaving()
//! Regression test for issue #1136 ("After a failed project saving, no saving takes place any more;
//! crash when changing projects")

TEST_F(TestSaveService, failingSaveService)
{
    const QString projectDir("test_failingSaveService");
    // do NOT create dir in order to force saving to fail
    const QString projectFileName(projectDir + "/document.pro");

    std::unique_ptr<ProjectDocument> document(new ProjectDocument);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");
    modify_models(*document);

    EXPECT_FALSE(QFile::exists(projectFileName));

    SaveService service;
    QSignalSpy spySaveService(&service, SIGNAL(projectSaved()));

    service.setDocument(document.get());
    EXPECT_THROW(service.save(projectFileName), Error);

    EXPECT_EQ(spySaveService.count(), 0);
    EXPECT_FALSE(QFile::exists(projectFileName));
    EXPECT_FALSE(service.isSaving());

    // after failed save, document should still be in modified state
    EXPECT_TRUE(document->isModified());
}

//! Testing SaveService on documents having nonXML data.
//! SaveService should be able to save project file (in main thread) and project nonXML
//! in second thread.

TEST_F(TestSaveService, saveServiceWithData)
{
    const QString projectDir("test_saveServiceWithData");
    UTest::GUI::create_dir(projectDir);
    const QString projectFileName(projectDir + "/document.pro");

    std::unique_ptr<ProjectDocument> document(new ProjectDocument);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");

    RealItem* realData = UTest::GUI::createRealData("TestData", *document->realModel());
    ASSERT(realData);
    DataItem* intensityItem = realData->dataItem();
    GUI::Model::JobItemUtils::createDefaultDetectorMap(
        intensityItem, document->collectedItems()->instrument2DItems().front());

    EXPECT_FALSE(QFile::exists(projectFileName));

    SaveService service;
    QSignalSpy spySaveService(&service, SIGNAL(projectSaved()));

    service.setDocument(document.get());
    service.save(projectFileName);

    spySaveService.wait(m_save_wait); // waiting saving in a thread is complete

    EXPECT_EQ(spySaveService.count(), 1);
    EXPECT_TRUE(QFile::exists(projectFileName));
    EXPECT_TRUE(QFile::exists(projectDir + "/realdata_TestData_0.int.gz"));
    EXPECT_FALSE(document->isModified());
}

//! Testing SaveService when autosave is enabled.

TEST_F(TestSaveService, autosaveEnabled)
{
    const QString projectDir("test_autosaveEnabled");
    UTest::GUI::create_dir(projectDir);
    const QString projectFileName(projectDir + "/document.pro");

    std::unique_ptr<ProjectDocument> document(new ProjectDocument());
    document->setProjectFileName(projectFileName);
    auto* instrument = document->collectedItems()->addInstrument<GISASInstrumentItem>();
    instrument->setInstrumentName("GISAS");

    RealItem* realData = UTest::GUI::createRealData("TestData", *document->realModel());
    DataItem* intensityItem = realData->dataItem();
    GUI::Model::JobItemUtils::createDefaultDetectorMap(
        intensityItem, document->collectedItems()->instrument2DItems().front());
    document->clearModified();
    EXPECT_FALSE(document->isModified());

    SaveService service;
    service.setAutosaveEnabled(true);
    const int autosave_time(200);
    service.setAutosaveTime(autosave_time);
    service.setDocument(document.get());

    QSignalSpy spySaveService(&service, SIGNAL(projectSaved()));
    service.save(projectFileName);

    spySaveService.wait(m_save_wait); // waiting saving in a thread is complete
    EXPECT_EQ(spySaveService.count(), 1);
    EXPECT_FALSE(document->isModified());
    EXPECT_TRUE(QFile::exists(projectDir + "/document.pro"));
    EXPECT_TRUE(QFile::exists(projectDir + "/realdata_TestData_0.int.gz"));
    spySaveService.clear();

    // modify several times and check SaveService signals
    for (size_t i = 0; i < 10; ++i)
        modify_models(*document);

    EXPECT_TRUE(document->isModified());

    spySaveService.wait(m_save_wait); // waiting saving in a thread is complete
    EXPECT_EQ(spySaveService.count(), 1);

    EXPECT_TRUE(QFile::exists(projectDir + "/autosave/document.pro"));
    EXPECT_TRUE(QFile::exists(projectDir + "/autosave/realdata_TestData_0.int.gz"));

    // after autosave the project has to be still in modified state
    EXPECT_TRUE(document->isModified());
    // after autosave, project file name should remain the same
    EXPECT_EQ(document->projectFileName(), projectFileName);
}
